/* this file was manually generated -- don't delete it */

INLINE void f_and PROT((int, int));
INLINE void f_and_eq PROT((int, int));
INLINE void f_compl PROT((int, int));
INLINE void f_dec PROT((int, int));
INLINE void f_div_eq PROT((int, int));
INLINE void f_end_catch PROT((int, int));
INLINE void f_eq PROT((int, int));
INLINE void f_lsh PROT((int, int));
INLINE void f_lsh_eq PROT((int, int));
INLINE void f_mod_eq PROT((int, int));
INLINE void f_mult_eq PROT((int, int));
INLINE void f_ne PROT((int, int));
INLINE void f_or PROT((int, int));
INLINE void f_or_eq PROT((int, int));
INLINE void f_parse_command PROT((int, int));
INLINE void f_range PROT((int, int));
INLINE void f_rsh PROT((int, int));
INLINE void f_rsh_eq PROT((int, int));
INLINE void f_subscript PROT((int, int));
INLINE void f_sub_eq PROT((int, int));
INLINE void f_switch PROT((int, int));
INLINE void f_xor PROT((int, int));
INLINE void f_xor_eq PROT((int, int));
INLINE void f_aggregate_assoc PROT((int, int));
INLINE void f_aggregate PROT((int, int));
INLINE void f_function_constructor PROT((int, int));
INLINE void f_function_split PROT((int, int));
INLINE void f_sscanf PROT((int, int));
